/*
 *			   Параметры модуля ядра mkopci.[ko/o]
 *
 *		 Модуль ядра может принимать параметры 'v' и 'omited'
 *  - 'v' принимает значение от 0 до 3 и влияет на количество информации выводимой
 *	в системный журнал.
 *	--- вырезка из svn-лога ---
 *	0 - comletely silent
 *	1 - (default) prints only detected BARs, reports loading/unloading
 *	2 - + prints IRQ and memory mapping events
 *	3 - + prints ioctl events
 *	---------------------------
 *
 *  - 'plx9050bug_quirk' управляет обходом ошибки контроллера PLX9050. Может принимать следующие значения:
 *	0 - обход ошибки отключен (по умолчанию)
 *	1 - обход ошибки включен
 *	для целей тестирования оборудования:
 *	2 - принудительное включение ошибки на BAR0 и BAR1
 *	3 - +принудительное включение ошибки на BAR0
 *	4 - +принудительное включение ошибки на BAR1
 *
 *  - 'omited' говорит драйверу, какое устройство не следует инициализировать при
 *	загрузке.
 *	Значением этого параметра для модуля ядра 2.4 является физический адрес устройства на шине,
 *	например "0x10800" без кавычек.
 *	В ядрах версий 2.6+ можно исключать сразу несколько устройств, передавая их
 *	значения через запятую, но не более 4-х.
 *
 *  Значения параметров можно указывать таким образом:
 *  $ insmod/modprobe mkopci.[ko/o] parameter1_name=value [ parameter2_name=value ]
 *
 * Запись физического адреса устройства в файл /proc/mkopci/core также управляет
 * "видимостью" устройства пользовательскими программами. Вернуть пропущеное устройство
 * можно командой 'echo ADDR > /proc/mkopci/core'. ADDR может быть либо простым номером
 * устройства, но обязательно в шестнадцатиричном виде, либо (для Linux-2.6) номером устройства в
 * стандартном формате Linux вида NM:XY.z.
 */

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
#include <linux/config.h>
#endif
#include <linux/module.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/uaccess.h>
#include <linux/pci.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/list.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 25)
#include <linux/moduleparam.h>
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
#include <linux/mutex.h>
#include <linux/interrupt.h>
#include <linux/cdev.h>
#else
#include <asm/semaphore.h>
#endif
#include <linux/spinlock.h>
#include <asm/uaccess.h>
#include <asm/io.h>

#include "mkopci.h"

#define MKO_VENDOR  0x6403
#define MKO_DEVICE1 0x0430
#define MKO_DEVICE2 0x0431
#define MKO_DEVICE3 0x0434

#define PLX9050BUG_BAR0    0x001
#define PLX9050BUG_BAR1    0x002
#define PLX9050BUG_INJECT  0x004

#if !defined(__user)
#define __user
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 8, 0)
static struct pci_device_id ids[] __devinitdata = {
#else
static struct pci_device_id ids[] = {
#endif
	{ MKO_VENDOR, MKO_DEVICE1, PCI_ANY_ID, PCI_ANY_ID, },
	{ MKO_VENDOR, MKO_DEVICE2, PCI_ANY_ID, PCI_ANY_ID, },
	{ MKO_VENDOR, MKO_DEVICE3, PCI_ANY_ID, PCI_ANY_ID, },
	{ 0, 0, }
};
MODULE_DEVICE_TABLE(pci, ids);

#ifdef __LP64__
#define PFMT	"llx"
#else
#define PFMT	"x"
#endif

static unsigned char drv_version = 0x11;
#define mko_pci_addr(bus, device, func, regoffs) (((bus&0xFF)<<16) | ((device&0x1F)<<11) | ((func&0x7)<<8) | (regoffs&0xFC))

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
static struct kmem_cache *mkopci_device_cache;
static dev_t devp;
static struct class *mkopci_class;
#else
static kmem_cache_t *mkopci_device_cache;
#endif
static struct rw_semaphore devices_sem;
static LIST_HEAD(devices);
static atomic_t devices_nr = ATOMIC_INIT(0);
static uint8_t req_bars;

// module parameters
static int plx9050bug_quirk;	// PLX9050 bug quirk
static int v = 1;	   // verbosity level
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 25)
module_param(plx9050bug_quirk, int, S_IRUGO);
module_param(v, int, S_IRUGO);
#else
MODULE_PARM(plx9050bug_quirk, "i");
MODULE_PARM(v, "i");
#endif
MODULE_PARM_DESC(plx9050bug_quirk, "PLX9050 bug quirk");
MODULE_PARM_DESC(v, "verbosity level");

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
static int omited[MAX_DEVICES_NR];  // only MAX_DEVICES_NR devices at once can be omited
static int omited_nr;
module_param_array(omited, int, &omited_nr, S_IRUGO);
MODULE_PARM_DESC(omited, "device(s) to omit");
#else   // 2.4 can't handle arrays as module parameters
static int omited;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 25)
module_param(omited, int, S_IRUGO);
#else
MODULE_PARM(omited, "i");
#endif
MODULE_PARM_DESC(omited, "device to omit");
#endif
#ifndef VM_RESERVED
//#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0)
#define VM_RESERVED (VM_DONTEXPAND | VM_DONTDUMP)
#endif
/***************************** Interrupt handling ****************************/
static int mkopci_wait_irq(struct mkopci_device *dev)
{
	unsigned long flags, LPCI_4C;
	unsigned long *lpci_4c_addr = (unsigned long *) (dev->core.lin_base[0] + 0x4C);

	spin_lock_irqsave(&dev->lock, flags);
	LPCI_4C = ioread32(lpci_4c_addr);
	iowrite32(LPCI_4C | 0x49, lpci_4c_addr);		   //  01001001
	spin_unlock_irqrestore(&dev->lock, flags);

	if (wait_event_interruptible(dev->wq, ioread32(lpci_4c_addr) & 0x24)) {
		return -EFAULT;
	}
	if (v > 1)
		pr_info("mkopci%d: irq recieved\n", dev->core.n_dev);

	spin_lock_irqsave(&dev->lock, flags);
	LPCI_4C = ioread32(lpci_4c_addr);
	iowrite32(LPCI_4C & ~0x40, lpci_4c_addr);		  //  ~01000000
	spin_unlock_irqrestore(&dev->lock, flags);

	return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 19)
static irqreturn_t mkopci_int_handler(int irq, void *dev)
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
static irqreturn_t mkopci_int_handler(int irq, void *dev, struct pt_regs *regs)
#else
static void mkopci_int_handler(int irq, void *dev, struct pt_regs *regs)
#endif
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	static irqreturn_t ret = IRQ_NONE;
#endif
	struct mkopci_device *mko_dev = (struct mkopci_device *)dev;
	unsigned long LPCI_4C, *lpci_4c_addr  = (unsigned long *) ((unsigned char *)(mko_dev->core.lin_base[0] + 0x4C));

	spin_lock(&mko_dev->lock);
	if ((LPCI_4C = ioread32(lpci_4c_addr)) & 0x24) {	  //   00100100
		iowrite32(LPCI_4C & ~0x40, lpci_4c_addr);	  //  ~01000000
		wake_up_interruptible(&mko_dev->wq);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
		ret = IRQ_HANDLED;
#endif
	}
	spin_unlock(&mko_dev->lock);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	return ret;
#endif
}

static int mkopci_request_irq(struct mkopci_device *dev)
{
	int ret = 0;

	if (dev->core.irq_requested) {
		pr_err("mkopci%d: irq already requested\n", dev->core.n_dev);
		return -EBUSY;
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 18)
	ret = request_irq(dev->core.irq, &mkopci_int_handler, IRQF_SHARED, dev->core.name, dev);
#else
	ret = request_irq(dev->core.irq, &mkopci_int_handler, SA_SHIRQ, dev->core.name, dev);
#endif
	if (ret) {
		pr_err("mkopci%d: failed to request irq %d\n", dev->core.n_dev, dev->core.irq);
		return ret;
	}

	dev->core.irq_requested++;

	if (strcmp(current->comm, "TestSetISR") && v > 1) {
		pr_info("mkopci%d: irq %d requested (%s)\n", dev->core.n_dev, dev->core.irq, current->comm);
	}

	return ret;
}

static void mkopci_free_irq(struct mkopci_device *dev)
{
	if (dev->core.irq_requested) {
		synchronize_irq(dev->core.irq);
		free_irq(dev->core.irq, dev);
		dev->core.irq_requested--;

		if (strcmp(current->comm, "TestSetISR") && v > 1) {
			pr_info("mkopci%d: irq %d released\n", dev->core.n_dev, dev->core.irq);
		}
	}
}
///////////////////////////////////////////////////////////////////////////////

/*************************** File operations *********************************/
static int mkopci_open(struct inode *inode, struct file *filp)
{
	int ret = 0;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	struct mkopci_device *dev = container_of(inode->i_cdev, struct mkopci_device, cdev);
#else
	struct mkopci_device *dev = NULL, *d = NULL;

	MOD_INC_USE_COUNT;

	list_for_each_entry(d, &devices, list) {
		if (d->core.n_dev == MINOR(inode->i_rdev)) {
			dev = d;
			break;
		}
	}
#endif

	if (dev == NULL) {
		return -ENODEV;
	}

	down_write(&dev->rwsem);
	if (dev->omited) {
		ret = -ENODEV;
		goto err;
	}
	if (!(filp->f_flags & O_NONBLOCK)) {
		if (dev->core.process) {
			ret = -EBUSY;
			goto err;
		}
		dev->backdoor = 0;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 39)
		dev->core.process = pid_nr(get_task_pid(current, PIDTYPE_PID));
#else
		dev->core.process = current->pid;
#endif
	} else {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 39)
		dev->backdoor = pid_nr(get_task_pid(current, PIDTYPE_PID));
#else
		dev->backdoor = current->pid;
#endif
		dev->core.process = 0;
	}

	filp->private_data = dev;

	if (strcmp(current->comm, "TstOpenClose") && v > 1)
		pr_info("mkopci%d: device opened in %s mode (%s)\n", dev->core.n_dev,
				dev->backdoor ? "backdoor" : "regular", current->comm);
	goto out;

err:
out:
	up_write(&dev->rwsem);
	return ret;
}

static int mkopci_release(struct inode *inode, struct file *filp)
{
	struct mkopci_device *dev;

	if (filp->private_data == NULL)
		return 0;

	dev = (struct mkopci_device *)filp->private_data;
	down_write(&dev->rwsem);
	dev->core.process = 0;
	dev->backdoor = 0;
	dev->core.c_bar = -1;
	filp->private_data = NULL;
	mkopci_free_irq(dev);

	if (strcmp(current->comm, "TstOpenClose") && v > 1)
		pr_info("mkopci%d: device released\n", dev->core.n_dev);
	up_write(&dev->rwsem);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
	MOD_DEC_USE_COUNT;
#else
#endif
	return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
static long mkopci_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
#else
static int mkopci_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
#endif
{
	int ret = 0, n = 0;
	struct mkopci_device *dev = (struct mkopci_device *)filp->private_data, *d;

	if (_IOC_TYPE(cmd) != MKO_IOC_MAGIC) {
		pr_err("mkopci%d: _IOC_TYPE(cmd) != MKO_IOC_MAGIC\n", dev->core.n_dev);
		return -ENOTTY;
	}

	if (_IOC_NR(cmd) > MKO_IOC_MAXNR) {
		pr_err("mkopci%d: _IOC_NR(cmd) > MKO_IOC_MAXNR\n", dev->core.n_dev);
		return -ENOTTY;
	}

	if (_IOC_DIR(cmd) & _IOC_READ) {
		ret = !access_ok(VERIFY_WRITE, (void __user *)arg, _IOC_SIZE(cmd));
	} else if (_IOC_DIR(cmd) & _IOC_WRITE) {
		ret = !access_ok(VERIFY_READ, (void __user *)arg, _IOC_SIZE(cmd));
	}

	if (ret) {
		pr_err("mkopci%d: mkopci_ioctl: access_ok failed\n", dev->core.n_dev);
		return -EIO;
	}

	switch (cmd) {
	case MKOPCI_IOCTL_CWPID:
		if (v > 2)
			pr_info("mkopci%d: MKOPCI_IOCTL_CWPID ioctl\n", dev->core.n_dev);
		down_read(&dev->rwsem);
		ret = dev->core.process;
		up_read(&dev->rwsem);
		break;
	case MKOPCI_IOCTL_GET_VERSION:
		if (v > 2)
			pr_info("mkopci%d: MKOPCI_IOCTL_GET_VERSION  ioctl\n", dev->core.n_dev);
		if (copy_to_user((void __user *)arg, &drv_version, sizeof(drv_version)))
			ret = -EFAULT;
		break;
	case MKOPCI_IOCTL_GET_BOARDS_COUNT:
		if (v > 2)
			pr_info("mkopci%d: MKOPCI_IOCTL_GET_BOARDS_COUNT ioctl\n", dev->core.n_dev);
		down_read(&dev->rwsem);
		n = atomic_read(&devices_nr);
		if (copy_to_user((void __user *)arg, &n, sizeof(n)))
			ret = -EFAULT;
		up_read(&dev->rwsem);
		break;
	case MKOPCI_IOCTL_GET_DEVICE_TABLE:
		if (v > 2)
			pr_info("mkopci%d: MKOPCI_IOCTL_GET_DEVICE_TABLE ioctl\n", dev->core.n_dev);
		down_read(&devices_sem);
		n = atomic_read(&devices_nr);
		if (copy_to_user((void __user *)arg, &n, sizeof(n))) {
			up_read(&devices_sem);
			ret = -EFAULT;
			break;
		}
		n = 0;
		list_for_each_entry(d, &devices, list) {
			down_read(&d->rwsem);
			if (d->omited) {
				up_read(&d->rwsem);
				continue;
			}
			if (copy_to_user((void __user *)arg + sizeof(unsigned short) + n * sizeof(struct mkopci_core),
						&d->core, sizeof(struct mkopci_core))) {
				up_read(&d->rwsem);
				ret = -EFAULT;
				break;
			}
			up_read(&d->rwsem);
			n++;
		}
		up_read(&devices_sem);
		break;
	case MKOPCI_IOCTL_ATTACH_IRQ:
		if (v > 2)
			pr_info("mkopci%d: MKOPCI_IOCTL_ATTACH_IRQ ioctl\n", dev->core.n_dev);
		down_write(&dev->rwsem);
		ret = mkopci_request_irq(dev);
		up_write(&dev->rwsem);
		break;
	case MKOPCI_IOCTL_DETACH_IRQ:
		if (v > 2)
			pr_info("mkopci%d: MKOPCI_IOCTL_DETACH_IRQ ioctl\n", dev->core.n_dev);
		down_write(&dev->rwsem);
		mkopci_free_irq(dev);
		up_write(&dev->rwsem);
		break;
	case MKOPCI_IOCTL_WAIT_IRQ:
		if (v > 2)
			pr_info("mkopci%d: MKOPCI_IOCTL_WAIT_IRQ ioctl\n", dev->core.n_dev);
		ret = mkopci_wait_irq(dev);
		break;
	case MKOPCI_IOCTL_REQUEST_BAR:
		if (v > 2)
			pr_info("mkopci%d: MKOPCI_IOCTL_REQUEST_BAR ioctl\n", dev->core.n_dev);
		down_write(&dev->rwsem);
		ret = get_user(dev->core.c_bar, (int __user *) arg);
		up_write(&dev->rwsem);
		break;
	default:
		pr_err("mkopci%d: invalid ioctl %d\n", dev->core.n_dev, _IOC_NR(cmd));
		ret = -EINVAL;
		break;
	};

	return ret;
}

#ifndef pgprot_noncached
static inline pgprot_t pgprot_noncached(pgprot_t _prot)
{
	unsigned long prot = pgprot_val(_prot);

	if (boot_cpu_data.x86 > 3)
		prot |= _PAGE_PCD | _PAGE_PWT;

	return __pgprot(prot);
}
#endif

static int mkopci_mmap(struct file *filp, struct vm_area_struct *vma)
{
	struct mkopci_device *dev = filp->private_data;
	unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;

	if (dev->core.c_bar == -1) {
		pr_err("mkopci%d: invalid c_bar number\n", dev->core.n_dev);
		return -EINVAL;
	}

	if (vma->vm_pgoff != 0) {
		pr_err("mkopci%d: vma->vm_pgoff != 0, aborting mapping\n", dev->core.n_dev);
		return -EINVAL;
	}

	if (PAGE_ALIGN(dev->core.mem_size[dev->core.c_bar]) != PAGE_ALIGN(vma->vm_end - vma->vm_start)) {
		pr_err("mkopci%d: PAGE_ALIGN error\n", dev->core.n_dev);
		return -EINVAL;
	}

	if (offset >= __pa(high_memory) || (filp->f_flags & O_SYNC))
		vma->vm_flags |= VM_IO;
	vma->vm_flags |= VM_RESERVED | VM_SHARED;

	if (v > 1 && strcmp("TstOpenClose", current->comm)) {
		pr_info("mkopci%d: .mmap BAR%d: [0x%lx - 0x%lx], vma [0x%lx - 0x%lx]\n",
			dev->core.n_dev, dev->core.c_bar, dev->core.mem_base[dev->core.c_bar],
			dev->core.mem_base[dev->core.c_bar] + dev->core.mem_size[dev->core.c_bar], vma->vm_start, vma->vm_end);
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 10)
	if (remap_pfn_range(vma, vma->vm_start,
							virt_to_phys(bus_to_virt(dev->core.mem_base[dev->core.c_bar])) >> PAGE_SHIFT,
							dev->core.mem_size[dev->core.c_bar], pgprot_noncached(vma->vm_page_prot)))  {
#else
#ifdef RH_RPR_VER
	if (remap_page_range(vma, vma->vm_start, dev->core.mem_base[dev->core.c_bar],
			vma->vm_end - vma->vm_start, pgprot_noncached(vma->vm_page_prot))) {
#else
	if (remap_page_range(vma->vm_start,
							dev->core.mem_base[dev->core.c_bar], vma->vm_end - vma->vm_start,
							pgprot_noncached(vma->vm_page_prot))) {
#endif
#endif
		pr_err("mkopci%d: memory remapping failed\n", dev->core.n_dev);
		return -EAGAIN;
	}

	return 0;
}

static const struct file_operations mkopci_fops = {
	.owner		  = THIS_MODULE,
	.open		   = mkopci_open,
	.release		= mkopci_release,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 11)
	.unlocked_ioctl = mkopci_ioctl,
#else
	.ioctl		  = mkopci_ioctl,
#endif
	.mmap		   = mkopci_mmap,
};
///////////////////////////////////////////////////////////////////////////////

/*************************** pci_driver functions ****************************/
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
static int mkopci_probe_cdevhelper(struct pci_dev *dev)
{
// actual addition of device node(s)
	struct mkopci_device *device = (struct mkopci_device *)pci_get_drvdata(dev);
	struct device *dev_;
	int err = 0;

	cdev_init(&device->cdev, &mkopci_fops);
	device->cdev.owner = THIS_MODULE;

	if ((err = cdev_add(&device->cdev, MKDEV(MAJOR(devp), MINOR(devp) + device->core.n_dev), 1))) {
		pr_err("mkopci%d: error while cdev_add\n", device->core.n_dev);
		return err;
	}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
	dev_ = device_create(mkopci_class, NULL, MKDEV(MAJOR(devp), MINOR(devp) + device->core.n_dev), NULL, device->core.name);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 18)
	dev_ = device_create(mkopci_class, NULL, MKDEV(MAJOR(devp), MINOR(devp) + device->core.n_dev), device->core.name);
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 18)
	if (IS_ERR(dev_)) {
		pr_err("mkopci%d: Unable to create device\n", device->core.n_dev);
		err = PTR_ERR(dev_);
		cdev_del(&device->cdev);
	}
#endif
	return err;
}
#endif

static int mkopci_plx9050workaround(struct pci_dev *dev, int plx9050bug)
{
	struct mkopci_device *device;
	int err = 0, bar;
	phys_addr_t plxphys = ~(phys_addr_t)0, oldphys;
	resource_size_t len;
	struct resource *res;

	device = (struct mkopci_device *)pci_get_drvdata(dev);
	if (!device) {
		err = -ENODEV;
		goto out;
	}

	for (bar = 0; bar <= 1; bar++) {
		if ((plx9050bug & PLX9050BUG_INJECT) || ((plx9050bug & (bar + 1)) == bar + 1)) {
			plxphys = ((oldphys = pci_resource_start(dev, bar)) & ~0xff) + ((plx9050bug & PLX9050BUG_INJECT) ? 0x80 : 0x0);
			if ((plx9050bug & PLX9050BUG_INJECT) && ((plx9050bug & (bar + 1)) == bar + 1)) {
				pr_warn("mkopci%d: [PLX9050 bug injection] BAR%d: 0x%"PFMT" -> 0x%"PFMT"\n", device->core.n_dev, bar, oldphys, plxphys);
				dev->resource[bar].start |= 0x80;
				dev->resource[bar].end = dev->resource[bar].start + 0x80 - 1;
				if (!(req_bars & (1 << bar))) {
					err = pci_request_region(dev, bar, device->core.name);
					if (err) {
						pr_err("mkopci%d: failed to request region %d\n", device->core.n_dev, bar);
						goto out;
					} else
						req_bars |= (1 << bar);
				}
			} else if (!(plx9050bug & PLX9050BUG_INJECT)) {
				res = &dev->resource[bar];
				if (!(pci_resource_start(dev, bar) & 0x80))
					continue;
				pr_warn("mkopci%d: [PLX9050 bug workaround] BAR%d: 0x%"PFMT" -> 0x%"PFMT"\n", device->core.n_dev, bar, oldphys, plxphys);
				res->start &= ~0xff;
				res->end = res->start + 0x80 - 1;
				if (!(req_bars & (1 << bar))) {
					err = pci_request_region(dev, bar, device->core.name);
					if (err) {
						err = allocate_resource(dev->resource[bar].parent, res, 0x80, res->parent->start,
							res->parent->end - 0x80, 0x100, NULL, NULL);
						if (err) {
							err = allocate_resource(&iomem_resource, res, 0x80, iomem_resource.start,
								iomem_resource.end - 0x80, 0x100, NULL, NULL);
						}
						if (err) {
							pr_err("mkopci%d: error allocating resourse for BAR%d\n", device->core.n_dev, bar);
							goto out;
						}
						if (!(req_bars & (1 << bar))) {
							err = pci_request_region(dev, bar, device->core.name);
							if (err) {
								pr_err("mkopci%d: failed to request region %d\n", device->core.n_dev, bar);
								goto out;
							} else
								req_bars |= (1 << bar);
						}
					} else
						req_bars |= (1 << bar);
				}
			} else {
				err = pci_request_region(dev, bar, device->core.name);
				if (err) {
					pr_err("mkopci%d: failed to request region %d\n", device->core.n_dev, bar);
					goto out;
				} else
					req_bars |= (1 << bar);
			}
			len = pci_resource_len(dev, bar);
			pci_write_config_dword(dev, bar == 0 ? PCI_BASE_ADDRESS_0 : PCI_BASE_ADDRESS_1, dev->resource[bar].start);

			device->core.mem_base[bar] = dev->resource[bar].start;
			device->core.mem_size[bar] = len;
		}
	}

out:
	return err;
}

static int mkopci_probe_helper(struct pci_dev *dev)
{
	int reg = 0, n = 0, err = 0, plx9050bug = 0;
	struct mkopci_device *device;

	device = (struct mkopci_device *)pci_get_drvdata(dev);
	if (!device)
		return -ENODEV;

	req_bars = 0;

	if (dev->device == MKO_DEVICE1 || plx9050bug_quirk >= 2 || plx9050bug_quirk == 1) {
		if (plx9050bug_quirk == 2) {
			if ((err = mkopci_plx9050workaround(dev, plx9050bug | PLX9050BUG_INJECT | PLX9050BUG_BAR0 | PLX9050BUG_BAR1)))
				goto fail;
		} else if (plx9050bug_quirk == 3) {
			if ((err = mkopci_plx9050workaround(dev, plx9050bug | PLX9050BUG_INJECT | PLX9050BUG_BAR0)))
				goto fail;
		} else if (plx9050bug_quirk == 4) {
			if ((err = mkopci_plx9050workaround(dev, plx9050bug | PLX9050BUG_INJECT | PLX9050BUG_BAR1)))
				goto fail;
		}

		if (pci_resource_start(dev, 0) & 0x80)
			plx9050bug = PLX9050BUG_BAR0;
		if (pci_resource_start(dev, 1) & 0x80)
			plx9050bug |= PLX9050BUG_BAR1;

		if (plx9050bug && plx9050bug_quirk == 1)
			if ((err = mkopci_plx9050workaround(dev, plx9050bug)))
				goto fail;
	}

	for (; reg < MAX_MEM_WIN + 1; reg++) {
		if (pci_resource_flags(dev, reg) & IORESOURCE_MEM) {
			device->core.mem_base[n] = pci_resource_start(dev, reg);
			if (!device->core.mem_base[n]) {
				pr_err("mkopci%d: Unable to get BAR%d\n", device->core.n_dev, n);
				err = -EFAULT;
				goto fail;
			}

			device->core.mem_size[n] = pci_resource_len(dev, reg);
			if (!device->core.mem_size[n]) {
				pr_err("mkopci%d: Unable to get size of BAR%d\n", device->core.n_dev, n);
				err = -EFAULT;
				goto fail;
			}

			if (!(req_bars & (1 << reg)) && (err = pci_request_region(dev, reg, device->core.name))) {
				pr_err("mkopci%d: couldn't request region 0x%x with size 0x%x (err = %d)\n", device->core.n_dev,
					(unsigned int)device->core.mem_base[n],
					(unsigned int)device->core.mem_size[n], err);
				iounmap((void *)device->core.lin_base[n]);
				goto fail;
			} else
				req_bars |= (1 << reg);

			device->core.lin_base[n] = (unsigned long)ioremap_nocache(device->core.mem_base[n], device->core.mem_size[n]);
			if (!device->core.lin_base[n]) {
				pr_err("mkopci%d: Unable to remap BAR%d[0x%lx:0x%lx]\n", device->core.n_dev, n, device->core.mem_base[n],
					device->core.mem_base[n] + device->core.mem_size[n] - 1);
				err = -EFAULT;
				goto fail;
			}
			if (v > 0)
				pr_info("mkopci%d: BAR%d = 0x%08lx, linear = 0x%08lx, len = 0x%x\n",
					device->core.n_dev, reg, device->core.mem_base[n], device->core.lin_base[n],
					device->core.mem_size[n]);
			n++;
		} else {
			if (v > 0)
				pr_info("mkopci%d: BAR%d = 0x%lx I/O region\n",
					device->core.n_dev, reg, (unsigned long) pci_resource_start(dev, reg));

			if (!(req_bars & (1 << reg)) && (err = pci_request_region(dev, reg, device->core.name))) {
				pr_err("mkopci%d: couldn't request I/O region (BAR%d)\n", device->core.n_dev, reg);
				goto fail;
			} else
				req_bars |= (1 << reg);
		}
	}
	device->core.mem_windows_nr = n;

	device->core.ltype = (*((unsigned short *)(device->core.lin_base[4]) + 3)) & 0xFF;
	device->core.irq = dev->irq;
	if (v > 0)
		pr_info("mkopci%d: irq = %d\n", device->core.n_dev, device->core.irq);
	device->core.irq_requested = 0;
	device->core.c_bar = -1;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	if ((err = mkopci_probe_cdevhelper(dev))) {
		plx9050bug = 0; // we need to unmap/release all bars
		goto fail;
	}
#endif
	atomic_inc(&devices_nr);
	goto out;

fail:
	for (reg = 0; reg < MAX_MEM_WIN + 1; reg++) {
		if (req_bars & (1 << reg)) {
			if (pci_resource_flags(dev, reg) & IORESOURCE_MEM) {
				iounmap((void *)device->core.lin_base[n]);
				pr_err("mkopci%d: unmapping reg%d\n", device->core.n_dev, n);
				device->core.lin_base[n--] = 0;
			}
			pci_release_region(dev, reg);
		}
	}

out:
	return err;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(3, 8, 0)
static int __devinit mkopci_probe(struct pci_dev *dev, const struct pci_device_id *id)
#else
static int mkopci_probe(struct pci_dev *dev, const struct pci_device_id *id)
#endif
{
	int n_prdev = 0, err = 0;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	int n;
#endif
	struct mkopci_device *device, *mdev = NULL;
	struct list_head *it;

	if (atomic_read(&devices_nr) == (1 << 8 * sizeof(unsigned char)) - 1)
		return -EBUSY;

	if ((err = pci_enable_device(dev))) {
		pr_err("mkopci: error = %d while enabling PCI device\n", err);
		return err;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 17)
	device = kmem_cache_zalloc(mkopci_device_cache, GFP_KERNEL);
#else
	device = kmem_cache_alloc(mkopci_device_cache, GFP_KERNEL);
#endif
	if (!device) {
		pr_err("mkopci: Unable to allocate memory\n");
		pci_disable_device(dev);
		return -ENOMEM;
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 17)
	memset(device, 0, sizeof(struct mkopci_device));
#endif

	down_write(&devices_sem);
	if (!list_empty(&devices)) {
		list_for_each(it, &devices) {
			mdev = list_entry(it, struct mkopci_device, list);
			if (n_prdev < mdev->core.n_dev) {
				if (n_prdev)
					device->core.n_dev = n_prdev - 1;
				else
					device->core.n_dev = n_prdev;
				list_add(&device->list, mdev->list.prev);
				break;
			}
			n_prdev++;
		}
		if (n_prdev >= mdev->core.n_dev) {
			device->core.n_dev = atomic_read(&devices_nr);
			list_add_tail(&device->list, &devices);
		}
	} else
		list_add_tail(&device->list, &devices);
	up_write(&devices_sem);

	pci_set_drvdata(dev, device);
	device->pci_dev = dev;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	for (n = 0; n < omited_nr; n++) {
		if (omited[n] == mko_pci_addr(dev->bus->number, PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn), 0)) {
#else
		if (omited == mko_pci_addr(dev->bus->number, PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn), 0)) {
#endif
			pci_disable_device(dev);
			device->omited = 1;
		}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	}
#endif

	device->core.vendor_id = dev->vendor;
	device->core.device_id = dev->device;
	device->core.subs_vendor_id = dev->subsystem_vendor;
	device->core.subs_id = dev->subsystem_device;
	device->core.instance = mko_pci_addr(dev->bus->number, PCI_SLOT(dev->devfn), PCI_FUNC(dev->devfn), 0);
	sprintf(device->core.name, "mkopci%d", device->core.n_dev);

	init_waitqueue_head(&device->wq);
	init_rwsem(&device->rwsem);
	spin_lock_init(&device->lock);

	down_write(&device->rwsem);
	if (device->omited)
		goto out;

	if ((err = mkopci_probe_helper(dev))) {
		pci_disable_device(dev);
		list_del(&device->list);
		kmem_cache_free(mkopci_device_cache, device);
	}

out:
	up_write(&device->rwsem);
	return err;
}

static void mkopci_remove_helper(struct pci_dev *dev)
{
	int reg, n = 0;
	unsigned char rom_base_reg = dev->rom_base_reg;
	struct mkopci_device *device = (struct mkopci_device *) pci_get_drvdata(dev);

	rom_base_reg = dev->rom_base_reg / 8;
	for (reg = 0; reg < DEVICE_COUNT_RESOURCE; reg++) {
		if (reg == rom_base_reg)
			continue;
		if (pci_resource_flags(dev, reg) & IORESOURCE_MEM) {
			if (v > 1)
				pr_info("mkopci%d: unmapping BAR%d (0x%lx)\n", device->core.n_dev, reg, device->core.lin_base[n]);
			iounmap((void *)device->core.lin_base[n]);
			device->core.lin_base[n++] = 0;
		}

		pci_release_region(dev, reg);
	}

	mkopci_free_irq(device);
	pci_disable_device(dev);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 18)
	device_destroy(mkopci_class, MKDEV(MAJOR(devp), MINOR(devp) + device->core.n_dev));
	cdev_del(&device->cdev);
#endif
	atomic_dec(&devices_nr);
}

static void mkopci_remove(struct pci_dev *dev)
{
	struct mkopci_device *device = (struct mkopci_device *) pci_get_drvdata(dev);

	down_write(&device->rwsem);
	if (!device->omited)
		mkopci_remove_helper(dev);
	list_del(&device->list);
	up_write(&device->rwsem);
	kmem_cache_free(mkopci_device_cache, device);
	if (v > 0)
		pr_info("mkopci%d: removed\n", device->core.n_dev);
}
///////////////////////////////////////////////////////////////////////////////

static struct pci_driver mkopci_driver = {
	.name       = "mkopci",
	.id_table   = ids,
	.probe      = mkopci_probe,
	.remove	    = mkopci_remove,
};

/*************************** proc entries functions **************************/
static struct proc_dir_entry *mkopci_proc_file, *mkopci_proc_dir, *mkopci_proc_core;

static int mkopci_proc_show(struct seq_file *m, void *vp)
{
	struct mkopci_device *mko_dev;

	down_read(&devices_sem);
	if (!atomic_read(&devices_nr))
		seq_puts(m, "no devices detected\n");
	else
		seq_printf(m, "%d device(s) detected\n", atomic_read(&devices_nr));

	seq_printf(m, "plx9050bug_quirk = %d\n", plx9050bug_quirk);
	seq_printf(m, "verbosity level = %d\n", v);

	list_for_each_entry(mko_dev, &devices, list) {
		down_read(&mko_dev->rwsem);
		seq_printf(m,
			"\ndevice\t\t\t/dev/%s\n"
			"vendor_id =\t\t0x%04x\n"
			"device_id =\t\t0x%04x\n"
			"subs_vendor_id =\t0x%04x\n"
			"subs_id =\t\t0x%04x\n"
			"ltype =\t\t\t%d\n"
			"instance =\t\t%02x:%02x.%x (0x%x)\n"
			"process =\t\t%d\n"
			"mem_base[0] =\t\t0x%08lx (linear = 0x%lx, size 0x%x)\n"
			"mem_base[1] =\t\t0x%08lx (linear = 0x%lx, size 0x%x)\n"
			"mem_base[2] =\t\t0x%08lx (linear = 0x%lx, size 0x%x)\n"
			"mem_base[3] =\t\t0x%08lx (linear = 0x%lx, size 0x%x)\n"
			"mem_base[4] =\t\t0x%08lx (linear = 0x%lx, size 0x%x)\n"
			"irq_n =\t\t\t%d\n"
			"irq_requested =\t\t%d\n",
					mko_dev->omited ? "(OMITED)" : mko_dev->core.name, mko_dev->core.vendor_id, mko_dev->core.device_id,
					mko_dev->core.subs_vendor_id, mko_dev->core.subs_id, mko_dev->core.ltype,
					mko_dev->core.instance >> 16,
					mko_dev->core.instance >> 11 & 0x001f,
					(mko_dev->core.instance >> 8) & 0x7, mko_dev->core.instance,
					mko_dev->core.process,
					mko_dev->core.mem_base[0], mko_dev->core.lin_base[0], mko_dev->core.mem_size[0],
					mko_dev->core.mem_base[1], mko_dev->core.lin_base[1], mko_dev->core.mem_size[1],
					mko_dev->core.mem_base[2], mko_dev->core.lin_base[2], mko_dev->core.mem_size[2],
					mko_dev->core.mem_base[3], mko_dev->core.lin_base[3], mko_dev->core.mem_size[3],
					mko_dev->core.mem_base[4], mko_dev->core.lin_base[4], mko_dev->core.mem_size[4],
					mko_dev->core.irq, mko_dev->core.irq_requested
				    );
		up_read(&mko_dev->rwsem);
	}
	up_read(&devices_sem);

	return 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 23)
static void *single_start(struct seq_file *m, loff_t *pos)
{
	return NULL + (*pos == 0);
}

static void *single_next(struct seq_file *m, void *v, loff_t *pos)
{
	++*pos;
	return NULL;
}

static void single_stop(struct seq_file *m, void *v)
{
}

static int single_open(struct file *file, int (*show)(struct seq_file *, void *), void *data)
{
	struct seq_operations *op = kmalloc(sizeof(*op), GFP_KERNEL);
	int res = -ENOMEM;

	if (op) {
		op->start = single_start;
		op->next = single_next;
		op->stop = single_stop;
		op->show = show;
		res = seq_open(file, op);
		if (!res)
			((struct seq_file *)file->private_data)->private = data;
		else
			kfree(op);
	}
	return res;
}

int single_release(struct inode *inode, struct file *file)
{
	struct seq_operations *op = ((struct seq_file *)file->private_data)->op;
	int res = seq_release(inode, file);

	kfree(op);
	return res;
}
#endif

static int proc_text_open(struct inode *inode, struct file *file)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	return single_open(file, mkopci_proc_show, inode->i_cdev);
#else
	return single_open(file, mkopci_proc_show, NULL);
#endif
}

static const struct file_operations mkopci_proc_fops = {
	.owner	 = THIS_MODULE,
	.open    = proc_text_open,
	.read    = seq_read,
	.llseek  = seq_lseek,
	.release = single_release,
};

static ssize_t mkopci_proc_core_read(struct file *file, char __user *buffer, size_t count, loff_t *offset)
{
	const int MAX_CORE_LEN = MAX_DEVICES_NR * sizeof(struct mkopci_core) + sizeof(unsigned short);
	int n, len = 0, ret = 0;
	struct mkopci_device *d;

	down_read(&devices_sem);
	n = atomic_read(&devices_nr) * sizeof(struct mkopci_core) + sizeof(unsigned short);
	if (*offset >= n)
		goto out;

	if (*offset == 0) {
		n = atomic_read(&devices_nr);
		if (copy_to_user(buffer, &n, sizeof(n))) {
			ret = -ERESTARTSYS;
			goto out;
		}
		len = sizeof(unsigned long);
	}

	list_for_each_entry(d, &devices, list) {
		down_read(&d->rwsem);
		if (d->omited) {
			up_read(&d->rwsem);
			continue;
		}
		if (copy_to_user(buffer + len, &d->core, sizeof(struct mkopci_core))) {
			up_read(&d->rwsem);
			ret = -ERESTARTSYS;
			goto out;
		}
		up_read(&d->rwsem);
		len += sizeof(struct mkopci_core);
	}

	if (count > len)
		if (len < MAX_CORE_LEN) {
			if (clear_user(buffer + len + 1, MAX_CORE_LEN - len)) {
				ret = -ERESTARTSYS;
				goto out;
			}
		}

	*offset += len;
	ret = len;

out:
	up_read(&devices_sem);
	return ret;
}

static int hex2int(const char s[])
{
	static const char hexalpha[] = "aAbBcCdDeEfF";
	unsigned int ret = 0;
	int i = 0, k;
	int err = 0;
	int hexint = 0;

	if (s[i] == '0') {
		i++;
		if (s[i] == 'x' || s[i] == 'X')
			i++;
	}

	while (!err && s[i] != '\0') {
		ret = ret << 4;
		if (s[i] >= '0' && s[i] <= '9')
			ret = ret + (s[i] - '0');
		else {
			for (k = 0; hexint == 0 && hexalpha[k] != '\0'; k++) {
				if (hexalpha[k] == s[i])
					hexint = 10 + k / 2;
			}
			if (hexint == 0) {
				err = -EINVAL;
				break;
			}
			ret = ret + hexint;
			hexint = 0;
		}
		i++;
	}

	if (err)
		ret = err;

	return ret;
}

static ssize_t mkopci_proc_core_write(struct file *file, const char __user *buffer, size_t count, loff_t *offset)
{
	char buf[8];
	int ret = 0, instance = 0, nodev = 1;
	struct mkopci_device *d;
	struct pci_dev *dev;

	if (copy_from_user(buf, buffer, min(sizeof(buf), count))) {
		ret = -EFAULT;
		goto out;
	}

	if (count > sizeof(buf)) {
		ret = -EINVAL;
		goto out;
	} else
		buf[count - 1] = 0;

	if (!strncmp(buf, "0x", 2) || !strncmp(buf, "0X", 2))
		instance = hex2int(buf);
	else
		if ((strlen(buf) == 7) && (buf[2] == ':') && (buf[5] == '.')) {
			int a, b, c;

			buf[2] = 0;
			buf[5] = 0;
			a = hex2int(buf);
			b = hex2int(buf+3);
			c = hex2int(buf+6);
			buf[2] = ':';
			buf[5] = '.';

			instance = mko_pci_addr(a, b, c, 0);
		} else
			instance = -EINVAL;

	if (instance == -EINVAL) {
		pr_err("mkopci: inappropriate PCI address '%s'\n", buf);
		ret = -EINVAL;
		goto out;
	}

	down_write(&devices_sem);
	list_for_each_entry(d, &devices, list) {
		down_write(&d->rwsem);
		if (d->core.instance != instance) {
			up_write(&d->rwsem);
			continue;
		}
		nodev = 0;
		if (d->core.process || d->backdoor) {
			if (v > 0) {
				if (d->core.process)
					pr_info("mkopci%d: device is handled by process [%d]\n", d->core.n_dev, d->core.process);
				else
					pr_info("mkopci%d: device is handled by process [%d] in backdoor mode\n", d->core.n_dev, d->backdoor);
			}
			ret = -EBUSY;
			goto out_up_sems;
		}
		dev = d->pci_dev;
		if (d->omited) {
			if ((ret = pci_enable_device(dev))) {
				pr_err("mkopci: error = %d while enabling PCI device\n", ret);
				goto out_up_sems;
			}

			if ((ret = mkopci_probe_helper(dev))) {
				pr_err("mkopci: error = %d in mkopci_probe_helper\n", ret);
				goto out_ph_fail;
			}
			d->omited = 0;
		} else {
			mkopci_remove_helper(dev);
			d->omited = 1;
		}
		up_write(&d->rwsem);
		break;
	}
	up_write(&devices_sem);

	if (nodev) {
		pr_info("mkopci: no device with address %s was found\n", buf);
		ret = -ENODEV;
		goto out;
	}
	ret = min(sizeof(buf), count);
	goto out;

out_ph_fail:
	pci_disable_device(dev);
out_up_sems:
	up_write(&d->rwsem);
	up_write(&devices_sem);
out:
	return ret;
}

static const struct file_operations mko_core_proc_fops = {
	.owner	  = THIS_MODULE,
	.read	   = mkopci_proc_core_read,
	.write	  = mkopci_proc_core_write,
};

static int mkopci_create_proc_entry(void)
{
#if LINUX_VERSION_CODE <= KERNEL_VERSION(3, 8, 13)
	mkopci_proc_dir = create_proc_entry("mkopci", S_IFDIR, NULL);
#else
	mkopci_proc_dir = proc_mkdir("mkopci", NULL);
#endif
	if (mkopci_proc_dir == NULL) {
		remove_proc_entry("mkopci", NULL);
		pr_err("mkopci: could not initialize /proc/mkopci/\n");
		return -ENOMEM;
	}

#if LINUX_VERSION_CODE <= KERNEL_VERSION(3, 8, 13)
	mkopci_proc_file = create_proc_entry("devices", S_IFREG | S_IRUGO, mkopci_proc_dir);
#else
	mkopci_proc_file = proc_create("devices", S_IFREG | S_IRUGO, mkopci_proc_dir, &mkopci_proc_fops);
#endif
	if (mkopci_proc_file == NULL) {
		remove_proc_entry("mkopci", NULL);
		pr_err("mkopci: could not initialize /proc/mkopci/devices\n");
		return -ENOMEM;
	}
#if LINUX_VERSION_CODE <= KERNEL_VERSION(3, 8, 13)
	mkopci_proc_file->proc_fops = &mkopci_proc_fops;

	mkopci_proc_core = create_proc_entry("core", S_IFREG | S_IRUGO | S_IWUSR | S_IWGRP, mkopci_proc_dir);
#else
	mkopci_proc_core = proc_create("core", S_IFREG | S_IRUGO | S_IWUSR | S_IWGRP, mkopci_proc_dir, &mko_core_proc_fops);
#endif
	if (mkopci_proc_core == NULL) {
		remove_proc_entry("devices", mkopci_proc_dir);
		remove_proc_entry("mkopci", NULL);
		pr_err("mkopci: could not initialize /proc/mkopci/core\n");
		return -ENOMEM;
	}

#if LINUX_VERSION_CODE <= KERNEL_VERSION(3, 8, 13)
	mkopci_proc_core->proc_fops = &mko_core_proc_fops;

	mkopci_proc_core->mode = S_IFREG | S_IRUGO | S_IWUSR | S_IWGRP;
	mkopci_proc_core->uid = 0;
	mkopci_proc_core->gid = 0;
	mkopci_proc_core->uid = 0;
#endif

	return 0;
}
///////////////////////////////////////////////////////////////////////////////
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
static int mkopci_major;
#endif

static int __init mkopci_init(void)
{
	int ret = 0;
	struct mkopci_device *d;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	if (v < 0 || v > 3 || omited_nr > MAX_DEVICES_NR) {
#else
	if (v < 0 || v > 3) {
#endif
		pr_err("mkopci: inappropriate verbosity parameters\n");
		ret = -EINVAL;
		goto out;
	}

	if (plx9050bug_quirk < 0 || plx9050bug_quirk > 4) {
		pr_err("mkopci: inappropriate plx9050bug_quirk parameters\n");
		ret = -EINVAL;
		goto out;
	}
	pr_info("mkopci: plx9050bug_quirk = %d\n", plx9050bug_quirk);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 23)
	mkopci_device_cache = kmem_cache_create("mkopci_dev_cache", sizeof(struct mkopci_device), 0,
								0, 0);
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
	EXPORT_NO_SYMBOLS;
#endif
	mkopci_device_cache = kmem_cache_create("mkopci_dev_cache", sizeof(struct mkopci_device), 0,
								SLAB_HWCACHE_ALIGN, NULL, NULL);
#endif

	if (!mkopci_device_cache) {
		pr_err("mkopci: Unable to allocate memory\n");
		ret = -ENOMEM;
		goto out;
	}
	init_rwsem(&devices_sem);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	mkopci_class = class_create(THIS_MODULE, "mkopci");
	if ((IS_ERR(mkopci_class))) {
		pr_err("mkopci: error creating class\n");
		ret = PTR_ERR(mkopci_class);
		goto destroy_cache;
	}

	if ((ret = alloc_chrdev_region(&devp, 0, MAX_DEVICES_NR, "mkopci"))) {
		pr_err("mkopci: failed to allocate chrdev region\n");
		goto destoy_class;
	}

	if (v > 0)
		pr_info("mkopci: major number = %d\n", MAJOR(devp));

	if ((ret = pci_register_driver(&mkopci_driver)) < 0) {
		pr_err("mkopci: error (%d) while pci_register_driver\n", ret);
		goto free_chr_reg;
	}
#else
	mkopci_major = register_chrdev(0, "mkopci", &mkopci_fops);
	if (mkopci_major < 0) {
		pr_info("mkopci: Unable to get major number\n");
		ret = mkopci_major;
		goto destroy_cache;
	}
	if (v > 0)
		pr_info("mkopci: mkopci_major = %d\n", mkopci_major);
	if ((ret = pci_register_driver(&mkopci_driver)) == 0) {
		pr_err("mkopci: no (%d) device was found, quiting\n", ret);
		goto free_chr_reg;
	}
#endif

	if ((ret = mkopci_create_proc_entry())) {
		pr_err("mkopci: error creating proc entries\n");
		goto unreg_drv;
	}
	goto out;

unreg_drv:
	pci_unregister_driver(&mkopci_driver);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
free_chr_reg:
	unregister_chrdev_region(devp, MAX_DEVICES_NR);
destoy_class:
	class_destroy(mkopci_class);
#else
free_chr_reg:
	unregister_chrdev(mkopci_major, "mkopci");
#endif
	while (!list_empty(&devices)) {
		d = list_entry(devices.next, struct mkopci_device, list);
		list_del(devices.next);
		kmem_cache_free(mkopci_device_cache, d);
	}
destroy_cache:
	kmem_cache_destroy(mkopci_device_cache);

out:
	return ret;
}

static void __exit mkopci_exit(void)
{
	struct mkopci_device *d;

	remove_proc_entry("core", mkopci_proc_dir);
	remove_proc_entry("devices", mkopci_proc_dir);
	remove_proc_entry("mkopci", NULL);

	pci_unregister_driver(&mkopci_driver);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	unregister_chrdev_region(devp, MAX_DEVICES_NR);
	class_destroy(mkopci_class);
#else
	unregister_chrdev(mkopci_major, "mkopci");
#endif
	while (!list_empty(&devices)) {
		d = list_entry(devices.next, struct mkopci_device, list);
		list_del(devices.next);
		kmem_cache_free(mkopci_device_cache, d);
	}
	kmem_cache_destroy(mkopci_device_cache);

	if (v > 0)
		pr_info("mkopci: exiting\n");
}

MODULE_DESCRIPTION("MKO PCI card driver");
MODULE_AUTHOR("Sergej Bauer");
MODULE_LICENSE("GPL");

module_init(mkopci_init);
module_exit(mkopci_exit);

